#![feature(proc_macro_diagnostic)]

use proc_macro::TokenStream;
use quote::quote;
use syn::{
    parse::{Parse, ParseStream},
    parse_macro_input, parse_quote,
    punctuated::Punctuated,
    spanned::Spanned,
    token::Comma,
    Expr, Item, ItemMod,
};

#[proc_macro_attribute]
pub fn day(args: TokenStream, input: TokenStream) -> TokenStream {
    let mut input = parse_macro_input!(input as ItemMod);
    let args = parse_macro_input!(args as Args);

    let content = if let Some(c) = &mut input.content {
        &mut c.1
    } else {
        input.span().unwrap().error("Module must have body!").emit();
        return quote!(#input).into();
    };

    let fns = content
        .iter()
        .filter_map(|i| match i {
            Item::Fn(fun) => Some(fun),
            _ => None,
        })
        .collect::<Vec<_>>();

    let part1 = fns
        .iter()
        .find(|i| &i.sig.ident.to_string() == "part1")
        .copied();
    let part2 = fns
        .iter()
        .find(|i| &i.sig.ident.to_string() == "part2")
        .copied();

    if part1.is_none() {
        input
            .span()
            .unwrap()
            .error("Need part1 function, which is not found within module.")
            .emit();
        return quote!(#input).into();
    }

    let part2 = if part2.is_some() {
        quote!(Some(self::part2))
    } else {
        quote!(None)
    };

    let name = args.name;
    content.push(parse_quote! {
        pub fn aoc_day() -> libaoc::Day {
            libaoc::Day {
                name: format!("{}", #name),
                part1: self::part1,
                part2: #part2,
            }
        }
    });

    if let Some(infile) = args.input {
        content.push(parse_quote!{
            const INPUT: &str = include_str!(#infile);
        });
    }

    quote!(#input).into()
}

struct Args {
    name: Expr,
    input: Option<Expr>,
}

impl Parse for Args {
    fn parse(input: ParseStream) -> syn::Result<Self> {
        let mut puncted = Punctuated::<Expr, Comma>::parse_separated_nonempty(input)?.into_iter();
        let name = puncted.next();
        if name.is_none() {
            return Err(syn::Error::new(
                input.span(),
                "Must provide at least one argument for the day name!",
            ));
        }

        let input = puncted.next();

        if let Some(excess) = puncted.next() {
            return Err(syn::Error::new(
                excess.span(),
                "No more than 2 arguments allowed!",
            ));
        }

        Ok(Self {
            name: name.unwrap(),
            input,
        })
    }
}
